// This file is part of crosci, licensed under the Academic Public License.
// See LICENSE.txt for more details.

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

double* dfa(double *seq, long npts, long *rs, int nr, double overlap_perc);